'''
Google Storage based File Transfer Implementation
'''
import urlparse

import errno
import sys
import os
import stat
import logging
import traceback
import time

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../.."))
from pilot.api import State
from bigjob import logger

from apiclient.discovery import build
from apiclient.http import MediaFileUpload
from oauth2client.file import Storage
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.client import Credentials
from oauth2client.tools import run
import httplib2
import urllib

"""
AN OAUTH2 Client Id must be created at the Google API console at:

https://code.google.com/apis/console/

=> API Access

More information with respect to OAUTH: https://developers.google.com/compute/docs/api/how-tos/authorization
"""
OAUTH2_CLIENT_ID='1004462711324-55akehip32m59u6omdfrt9s8u8ehb0hm.apps.googleusercontent.com'
OAUTH2_CLIENT_SECRET='EIMML1W7anu0XijVghws0DY-'

GS_PROJECT_ID="1004462711324"


class GSFileAdaptor(object):
    """ BigData File Management for Pilot Data """
   
    
    def __init__(self, service_url, security_context=None):        
        # Initializations of instance variables
        self.service_url = service_url
        self.bucket_name = self.__get_bucket_name(service_url)
        self.__state=State.New
                
        # Do OAUTH authentication
        if security_context!=None:
            logger.debug("Attempt to restore credentials from security context: " + str(security_context))
            self.credentials = Credentials.new_from_json(security_context)
        else:
            storage = Storage('gce.dat')
            self.credentials = storage.get()
        if self.credentials is None or self.credentials.invalid == True:
            logger.debug("No valid credential found. Run new OAuth authentication round...")
            flow = OAuth2WebServerFlow(
                                       client_id=OAUTH2_CLIENT_ID,
                                       client_secret=OAUTH2_CLIENT_SECRET,
                                       scope=['https://www.googleapis.com/auth/devstorage.full_control', 
                                              'https://www.googleapis.com/auth/compute'],
                                       user_agent='bigjob-client/1.0')

            self.credentials = run(flow, storage)

    
    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        return self.credentials.to_json()
    
        
    def initialize_pilotdata(self):
        # check whether directory exists
        self.__state=State.Running
        request_dict = {
                        "id":self.bucket_name,
                        "projectId":GS_PROJECT_ID
        }       
        logger.debug(str(request_dict)) 
        try:
            gs = self.__get_api_client()[0]
            gs.buckets().insert(body=request_dict).execute()
        except:
            logger.debug("Error creating bucket: " + self.bucket_name)
            pass # Do nothing if bucket already exists
                
        
    def get_pilotdata_size(self):
        # unlimited size
        return None
    
    
    def delete_pilotdata(self):
        self.__state=State.Done
    
        
    def get_state(self):
        return self.__state
        
            
    def create_du(self, du_id):
        gs = self.__get_api_client()[0]
        o = gs.objects().insert(bucket=self.bucket_name, name=str(du_id)+"/du_info",
                                body={'media': {
                                                "contentType":"text/ascii",
                                                "data": du_id    
                                                } 
                                      }                                
                                ).execute()
        logger.debug("Created GS: " + str(o))
        
                
    def put_du(self, du):
        logger.debug("Copy DU to Google Storage")
        du_items = du.list()
        for i in du_items.keys():   
            try:
                local_filename=du_items[i]["local"]
                remote_path = os.path.join(str(du.id), os.path.basename(local_filename))
                self._put_file(local_filename, remote_path)
            except:
                logger.debug("Could not copy file: " + (str(i)))
            
    
    def copy_du(self, du, pd_new):
        pass
        
    
    def get_du(self, du, target_url):
        du_id=du.id
        logger.debug("Get DU: " + str(du_id))
        gs = self.__get_api_client()[0]
        result = gs.objects().list(bucket=self.bucket_name, prefix=du_id).execute() 
                            #delimiter="/",
                            #prefix=[du_id]).execute()   
        logger.debug("Result: " + str(result))
        for i in result["items"]:
            full_filename = i["name"]
            self._get_file(full_filename, os.path.join(target_url, os.path.basename(full_filename)))
        
                
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.bucket_name, du.id))
    
    
    ###########################################################################
    # Pure File Management APIs
    def _put_file(self, source, target):
        logger.debug("Put file: %s to %s"%(source, target))
        gs = self.__get_api_client()[0]
        #media = MediaFileUpload(source, 
        #                        resumable=False)
        o = gs.objects().insert(bucket=self.bucket_name, 
                                name=target,
                                media_body=source).execute()            
        logger.debug("Put file result: %s"%str(o))
    
    
    def _get_file(self, source, target):
        logger.debug("GET file: %s to %s"%(source, target))
        gs, http = self.__get_api_client()
        f = gs.objects().get(bucket=self.bucket_name, 
                             object=source).execute()            
        logger.debug("Get file result: %s"%str(f))
        downloadUrl = f["media"]['link']
        if downloadUrl:
            response, content = http.request(downloadUrl)
            logger.debug("Download file response: %d"%(response.status))
            with open(target, 'wb') as f:
                f.write(content)
        
    
    def transfer(self, source_url, target_url):
        pass
    
    def create_remote_directory(self, target_url):
        return True
    
    
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        target_path = result.path
        return target_path
        
        
    ###########################################################################
    # Auxiliary functions
    def __get_api_client(self):
        http = httplib2.Http()
        http = self.credentials.authorize(http)
        gs = build("storage", "v1beta1", http=http)
        return gs, http
    
    
    def __get_bucket_name(self, service_url):
        bucket_name = service_url.replace("gs://", "")
        bucket_name = bucket_name.replace("/", "")
        return bucket_name
        
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
    
if __name__ == "__main__":
    gs = GSFileAdaptor("gs://pilot-data-bucket-1234")
    gs.initialize_pilotdata()
    gs._put_file("test-random.exe", "test.exe")
    gs._get_file("test.txt", "test2.txt")
    gs.get_du(None, ".")


import sys
import os
import time
import logging
logging.basicConfig(level=logging.DEBUG)

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, os.getcwd() + "/../")
from pilot import PilotComputeService, ComputeDataService, State

COORDINATION_URL = "redis://localhost:6379"

if __name__ == "__main__":      
    
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description = {
                             "service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             "working_directory": os.path.join(os.getcwd(),"work"),
                             'affinity_datacenter_label': "eu-de-south",              
                             'affinity_machine_label': "mymachine" 
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
    pilotjob2 = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
         
    compute_data_service = ComputeDataService()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [""],
            "total_core_count": 1,
            "number_of_processes": 1,            
            "output": "stdout.txt",
            "error": "stderr.txt",   
            "affinity_datacenter_label": "eu-de-south",              
            "affinity_machine_label": "mymachine" 
    }
    
    cus = []
    for i in range(0,10):
        compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
        cus.append(compute_unit)
    
    
    logging.debug("Finished setup. Waiting for scheduling of CU")
    compute_data_service.wait()

    
    logging.debug("Terminate Pilot Compute and Compute Data Service")
    compute_data_service.cancel()    
    pilot_compute_service.cancel()



import sys
import os
import time
import logging
import uuid
#logging.basicConfig(level=logging.DEBUG)

#sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotComputeService, PilotDataService, ComputeDataService, State
from bigjob import logger 

#COORDINATION_URL = "redis://localhost:6379"
COORDINATION_URL="redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379/pcs/pcs-4867ff08-e192-11e1-a694-00003e980000"

if __name__ == "__main__":      
     
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description_amazon = {
                             "service_url": 'ec2+ssh://aws.amazon.com',
                             #"service_url": 'euca+ssh://198.202.120.90/',
                             #"service_url": 'gce+ssh://api.google.com',
                             #"service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-google",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             "vm_id":"ami-d7f742be",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"MyKey",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/id_rsa",
                             "vm_type":"t1.micro",
                             "access_key_id":"x",
                             "secret_access_key":"x"
                            }
    
    pilot_compute_description_amazon_west = {
                             "service_url": 'ec2+ssh://aws.amazon.com',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-google",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             #"vm_id":"ami-d7f742be",
                             "vm_id": "ami-7e37173b",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"MyKey",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/id_rsa",
                             "vm_type":"t1.micro",
                             "region" : "us-west-1",
                             "access_key_id":"x",
                             "secret_access_key":"x"
                           
                            }
    
    pilot_compute_description_euca_sierra = {
                             #"service_url": 'ec2+ssh://aws.amazon.com',
                             "service_url": 'euca+ssh://198.202.120.90/',
                             #"service_url": 'gce+ssh://api.google.com',
                             #"service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-west",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             "vm_id":"ami-d7f742be",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"MyKey",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/id_rsa",
                             "vm_type":"c1.xlarge",
                             "access_key_id":"OQTYSHQ9I6ACEXDTKTOGG",
                             "secret_access_key":"STgYQPTyh89EWKp7lFtd77A4O2aghXjx0owqkNHm"
                            }
    
    pilot_compute_description_euca_india = {
                             "service_url": 'euca+ssh://149.165.146.135:8773/services/Eucalyptus',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-east",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             "vm_id":"emi-9DCC3DFA",
                             "vm_ssh_username":"root",
                             "vm_ssh_keyname":"luckow",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/eucakey-india",
                             "vm_type":"c1.xlarge",
                             "access_key_id":"8MCXRAMXMHDYKWNKXZ8WF",
                             "secret_access_key":"YrcUqSw2Arxshrh3ZtenkxerWwCWdMTKvZYoLPAo"
                            }
    
    pilot_compute_description_openstack_india = {
                             "service_url": 'nova+ssh://149.165.146.50:8773/services/Cloud',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-east",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             "vm_id":"ami-00000046",
                             "vm_ssh_username":"root",
                             "vm_ssh_keyname":"novakey-india",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/novakey-india",
                             "vm_type":"m1.xlarge",
                             "access_key_id":"f9716a49c92a4a4cbedb6aba5e78d682",
                             "secret_access_key":"bcdff54b7fe94d63b4412c762e823a84"
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description_amazon_west)
    
    compute_data_service = ComputeDataService()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    
    
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [],
            "number_of_processes": 1,
            "output": "stdout.txt",
            "error": "stderr.txt",   
            "input_data": [],
            "output_data": []
    }    
    
    compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
    logging.info("Finished setup of ComputeDataService. Waiting for scheduling of PD")
    compute_data_service.wait()
    
    
    logging.info("Terminate Pilot Compute/Data Service")
    compute_data_service.cancel()
    pilot_compute_service.cancel()


#!/usr/bin/env python

from bigjob import logger
from apiclient.discovery import build
from oauth2client.file import Storage
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.tools import run
import httplib2
import os
import uuid
import time

import bliss.saga as saga

"""
AN OAUTH2 Client Id must be created at the Google API console at:

https://code.google.com/apis/console/

=> API Access

More information with respect to OAUTH: https://developers.google.com/compute/docs/api/how-tos/authorization
"""
OAUTH2_CLIENT_ID='1004462711324-55akehip32m59u6omdfrt9s8u8ehb0hm.apps.googleusercontent.com'
OAUTH2_CLIENT_SECRET='EIMML1W7anu0XijVghws0DY-'

GCE_PROJECT_ID='bigjob-pilot'

"""
Google Compute Engine currently provides a default image with Ubuntu 12.04

To use BJ, a custom image containing gcc and build essentials needs to be
created.

$ apt-get update
$ apt-get install gcc python-all-dev 

"""
GCE_IMAGE_URL="https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/images/bigjob-image"

class gce_states:
    PROVISIONING="PROVISIONING"
    STAGING="STAGING"
    RUNNING="RUNNING"
    
    
class Service(object):
    """ Plugin for Google Compute Engine 
    
        Manages endpoint in the form of:
        
            gce+ssh://api.google.com    
    """

    def __init__(self, resource_url, pilot_compute_description):
        """Constructor"""
        self.resource_url = resource_url
        self.pilot_compute_description =  pilot_compute_description   
    
    def create_job(self, job_description):
        j = Job(job_description, self.resource_url, self.pilot_compute_description)
        return j
            
    
    def __del__(self):
        pass
    
    
    

class Job(object):


    def __init__(self, job_description, saga_url, pilot_compute_description):
        
        self.job_description = job_description
        self.saga_url = saga_url
        self.pilot_compute_description = pilot_compute_description
        self.image_url = GCE_IMAGE_URL
        if self.pilot_compute_description.has_key("vm_id"):
            self.image_url = self.pilot_compute_description["vm_id"]
            
        self.machine_type = "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/machine-types/n1-standard-1"
        if self.pilot_compute_description.has_key("vm_type"):
            self.machine_type = self.pilot_compute_description["vm_type"]
        
        self.location = "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/zones/us-east1-a"
        if self.pilot_compute_description.has_key("vm_location"):
            self.location = self.pilot_compute_description["vm_location"]
       
            
        self.id="bigjob-" + str(uuid.uuid1())
        self.network_ip=None
        
        # Do OAUTH authentication
        storage = Storage('gce.dat')
        self.credentials = storage.get()
        if self.credentials is None or self.credentials.invalid == True:
            flow = OAuth2WebServerFlow(
                                       client_id=OAUTH2_CLIENT_ID,
                                       client_secret=OAUTH2_CLIENT_SECRET,
                                       scope='https://www.googleapis.com/auth/compute',
                                       user_agent='bigjob-client/1.0')

            self.credentials = run(flow, storage)

        
        
    def run(self):
        request_dict = {
          "kind": "compute#instance",
          "disks": [
                    {
                     "kind": "compute#instanceDisk",
                     "type": "PERSISTENT",
                     "mode": "READ",
                     "deviceName": "reference-genome",
                     "source": "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/disks/reference-genome"
                     }
          ],
          "networkInterfaces": [
            {
              "kind": "compute#instanceNetworkInterface",
              "accessConfigs": [
                {
                  "name": "External NAT",
                  "type": "ONE_TO_ONE_NAT"
                }
              ],
              "network": "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/networks/default"
            }
          ],
          "serviceAccounts": [
                              {
                               "kind": "compute#serviceAccount",
                               "email": "default",
                               "scopes": [
                                          "https://www.googleapis.com/auth/userinfo.email",
                                          "https://www.googleapis.com/auth/compute",
                                          "https://www.googleapis.com/auth/devstorage.full_control"
                                          ]
                               }
                              ],         
          #"zone": "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/zones/us-east1-a",
          "zone": self.location,
          #"machineType": "https://www.googleapis.com/compute/v1beta12/projects/bigjob-pilot/machine-types/n1-standard-1",
          "machineType": self.machine_type,
          "name": self.id,
          "image": self.image_url       
        }
         
        http = httplib2.Http()
        http = self.credentials.authorize(http)
        gce = build("compute", "v1beta12", http=http)
        #result = gce.instances().get(instance="bigjob-pilot", project="bigjob-pilot").execute()
        gce.instances().insert(project=GCE_PROJECT_ID, body=request_dict).execute()
        
        time.sleep(15) # wait for startup
        #wait for compute instance to become active
        self.wait_for_running()
        
        # spawn BJ agent via SSH
        compute_instance_details = self.__get_instance_resource()
        logger.debug("Compute Instance Details: " + str(compute_instance_details))
        self.network_ip = compute_instance_details["networkInterfaces"][0]["accessConfigs"][0]['natIP']
        url = "ssh://" + str(self.network_ip)
        logger.debug("Connect to: %s"%(url))
        js = saga.job.Service(url)
        
        # Submit job
        ctx = saga.Context()
        ctx.type = saga.Context.SSH
        ctx.userid  = self.pilot_compute_description["vm_ssh_username"]
        ctx.userkey = self.pilot_compute_description["vm_ssh_keyfile"]
        js.session.contexts = [ctx]


        job = js.create_job(self.job_description)
        print "Submit pilot job to: " + str(url)
        
        TRIAL_MAX=15
        trials=0
        while trials < TRIAL_MAX:
            try:
                logger.debug("Attempt: %d, submit pilot job to: %s "%(trials,str(url)))
                job.run()
                break
            except:
                trials = trials + 1 
                time.sleep(10)
                if trials == TRIAL_MAX:
                    raise Exception("Submission of agent failed.") 
                
        logger.debug("Job State : %s" % (job.get_state())) 
        
      
        
        print "Job State : %s" % (job.get_state())
       

    def wait_for_running(self):
        while self.get_state()!=gce_states.RUNNING:
            time.sleep(5)
        
    
    def get_state(self):
        result=self.__get_instance_resource()
        return result["status"]
    
    
    def cancel(self):
        http = httplib2.Http()
        http = self.credentials.authorize(http)
        gce = build("compute", "v1beta12", http=http)
        gce.instances().delete(project=GCE_PROJECT_ID, instance=self.id).execute()
        
    
    def __get_instance_resource(self):
        http = httplib2.Http()
        http = self.credentials.authorize(http)
        gce = build("compute", "v1beta12", http=http)
        result = gce.instances().get(project=GCE_PROJECT_ID, instance=self.id).execute()
        return result
    
    
 
     
if __name__ == "__main__":
    gce_service = Service("gce+ssh://api.google.com")
    j = gce_service.create_job(job_description)
    
    gce.run()
    print gce.get_state()



""" Affinity-aware scheduler that evaluates affinity labels and input/output data flow
    

"""
import random
import logging
from bigjob import logger

class Scheduler:
    
    def __init__(self):
        self.pilot_data=[]
        self.pilot_jobs=[]
    
    def set_pilot_data(self, pilot_data):
        """ set resources which are used for scheduling """
        self.pilot_data=pilot_data
        
    
    def set_pilot_jobs(self, pilot_jobs):
        """ set resources which are used for scheduling """
        self.pilot_jobs=pilot_jobs
    
        
    def schedule_pilot_data(self, data_unit_description=None):
        logger.debug("Schedule to PD - # Avail pilots: %d"%len(self.pilot_data))     
        candidate_pilot_data = []  
        if data_unit_description.has_key("affinity_datacenter_label") and data_unit_description.has_key("affinity_machine_label"):
            for i in self.pilot_data: 
                pilot_data_description = i.pilot_data_description
                if pilot_data_description.has_key("affinity_machine_label") and pilot_data_description.has_key("affinity_datacenter_label"):
                    if data_unit_description["affinity_datacenter_label"] == pilot_data_description["affinity_datacenter_label"]\
                    and data_unit_description["affinity_machine_label"] == pilot_data_description["affinity_machine_label"]:
                        candidate_pilot_data.append(i)
        
        if len(candidate_pilot_data) == 0:
            # No PD with requested affinity found
            # move data unit into a "possibly" remote pilot data
            logger.debug("No pilot data w/ affinity found... Looking for alternative pilot.")
            candidate_pilot_data = self.pilot_data
            
        if len(candidate_pilot_data)>0:
            return random.choice(candidate_pilot_data)
        
        return None
        
        #if len(self.pilot_data)!=0:
        #    return random.choice(self.pilot_data)
        return None
    
    
    def schedule_pilot_job(self, compute_unit_description=None):
        """ Enforces affinity description: if no PJ is available with the right
            affinity, WU can't be scheduled.
            
            TODO: incorporate potential data movements to co-locate PD/WU 
        
        """    
        logger.debug("Schedule to PJ - # Avail PJs: %d"%len(self.pilot_jobs))
        candidate_pilot_jobs = []
        required_number_of_processes=1 
        if compute_unit_description.has_key("number_of_processes"):
            required_number_of_processes = int(compute_unit_description["number_of_processes"])
        
        if compute_unit_description.has_key("affinity_datacenter_label") and compute_unit_description.has_key("affinity_machine_label"):
            for i in self.pilot_jobs:
                free_nodes = i.get_free_nodes()
                logger.debug("BJ: %r State: %s Free nodes: %d"%(i, i.get_state(), free_nodes))
                if i.get_state()=="Running" and free_nodes >= required_number_of_processes: # check whether pilot is active
                    pilot_job_description = i.pilot_compute_description
                    if pilot_job_description["affinity_datacenter_label"] == compute_unit_description["affinity_datacenter_label"]\
                        and pilot_job_description["affinity_machine_label"] == compute_unit_description["affinity_machine_label"]:
                        candidate_pilot_jobs.append(i)
                        
                        
        if len(candidate_pilot_jobs) == 0:
            # No PJ with requested affinity found
            # move compute unit into a "possibly" remote pilot job
            logger.debug("No pilot compute w/ affinity found... Looking for alternative pilot.")
            for i in self.pilot_jobs:                
                logger.debug("BJ: %r State: %s"%(i, i.get_state()))
                free_nodes = i.get_free_nodes()
                if i.get_state()=="Running" and free_nodes >= required_number_of_processes:
                    candidate_pilot_jobs.append(i)
                    #candidate_pilot_jobs=self.pilot_jobs

                
        logger.debug("Candidate PJs: %r"%(candidate_pilot_jobs))   
        if len(candidate_pilot_jobs)>0:
            return random.choice(candidate_pilot_jobs)
        
        return None
    
    def __check_pilot_data_dependency(self, work_unit_description):
        pilot_data_dependencies = work_unit_description["input_pilot_data"]
        for i in pilot_data_dependencies:
            pd = PilotData.pilot
            ps = i.get_pilot_data()



import sys
import os
import time
import logging
from pilot import PilotComputeService, ComputeDataService, State
logging.basicConfig(level=logging.WARNING)

COORDINATION_URL = "redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379"

if __name__ == "__main__":      
    
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)
    pilot_compute_description=[]

    # create pilot job service and initiate a pilot job
    pilot_compute_description.append({ "service_url": 'sge-ssh://pmantha@login1.ls4.tacc.utexas.edu',
                                  "number_of_processes":24,                             
                                  "walltime":10,
				  "processes_per_node":12,
                                  "queue":"normal", 
                                  "allocation":"TG-MCB090174",
                                  "working_directory": "/home1/01539/pmantha/agent",
                                })
    
    pilot_compute_description.append({
                             "service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             "working_directory": os.path.join(os.getcwd(),"work"),
                            })

    pilot_compute_description.append({ "service_url": 'sge+ssh://tg804093@login3.ranger.tacc.utexas.edu',
                                  "number_of_processes":16,
                                  "walltime":10,
                                  "processes_per_node":16,
                                  "queue":"normal",
                                  "allocation":"TG-MCB090174",
                                  "working_directory": "/work/01131/tg804093"
                                })

    pilot_compute_description.append({ "service_url": 'pbs-ssh://pmantha@kraken-gsi.nics.teragrid.org',
                                  "number_of_processes":12,
                                  "walltime":10,
                                  "processes_per_node":1,
                                  "queue":"small",
                                  "allocation":"TG-MCB090174",
                                  "working_directory": "/lustre/scratch/pmantha/agent/",
                                })
    #for pcd in pilot_compute_description:
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description[2])
         
    compute_data_service = ComputeDataService()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [""],
            "total_core_count": 1,
            "number_of_processes": 1,            
            "output": "stdout.txt",
            "error": "stderr.txt"
    }
    
    for i in range(0,20):
        compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
    
    
    logging.debug("Finished setup. Waiting for scheduling of CU")
    compute_data_service.wait()

    
    logging.debug("Terminate Pilot Compute and Compute Data Service")
    compute_data_service.cancel()    
    pilot_compute_service.cancel()



'''
Encapsulates coordination and communication specifics of bigjob
'''
import logging
import threading
import datetime
import sys
import os
import pickle
import pdb
import time

from bigjob import logger
import redis

    
logging.debug(str(sys.path))
import uuid


REDIS_SERVER="localhost"
REDIS_SERVER_PORT=6379
REDIS_URL_SCHEME="redis://"

class bigjob_coordination(object):
    '''
    Encapsulates communication and coordination
    Implementation based on Redis (http://redis.io)
    '''

    def __init__(self, server=REDIS_SERVER, server_port=REDIS_SERVER_PORT, server_connect_url=None,
                 username=None, password=None, dbtype=None, url_prefix=None):
        '''
        Constructor
        '''
        if server_port==None:
            server_port=6379
            
        self.username = None
        self.password = None   
        
        self.address = "%s%s:%i"%(REDIS_URL_SCHEME, server, server_port)
        self.dbtype=""
        if server_connect_url!=None:
            self.address=server_connect_url    
            start_index = self.address.find(REDIS_URL_SCHEME)+len(REDIS_URL_SCHEME)
            server_and_port = self.address[start_index:]
            password_end = server_and_port.find("@") 
            # parse out password
            if password_end != -1:
                self.password = server_and_port[:password_end]
                start_index=password_end
                server_and_port= server_and_port[(password_end+1):]
            
            # port and hostname
            if server_and_port.find(":")==-1:
                server=server_and_port
                server_port = REDIS_SERVER_PORT
            else:
                server = server_and_port.split(":")[0]
                server_port = int(server_and_port.split(":")[1])
        else:
            self.password = username
            if self.password != None and self.password!="":
                self.address = "%s%s@%s:%i"%(REDIS_URL_SCHEME, self.password, server, server_port)
        
        logger.debug("Connect to Redis: " + server + " Port: " + str(server_port))
        
        if self.password==None:
            self.redis_client = redis.Redis(host=server, port=server_port, db=0)
        else:
            self.redis_client = redis.Redis(host=server, port=server_port, password=self.password, db=0)
        #self.redis_client_pubsub = self.redis_client.pubsub() # redis pubsub client       
        #self.resource_lock = threading.RLock()
        self.pipe = self.redis_client.pipeline()
        try:
            self.redis_client.ping()
        except:
            logger.error("Please start Redis server!")
            raise Exception("Please start Redis server!")
        
        
    def get_address(self):
        return self.address
    
    #####################################################################################
    # Pilot-Job State
    def set_pilot_state(self, pilot_url, new_state, stopped=False):     
        logger.debug("update state of pilot job to: " + str(new_state) 
                     + " stopped: " + str(stopped))
        self.redis_client.hmset(pilot_url, {"state":str(new_state), "stopped":str(stopped)})
        if stopped==True:
            self.queue_job(pilot_url, "STOP")
        
        
    def get_pilot_state(self, pilot_url):
        state = self.redis_client.hgetall(pilot_url)
        return state
    
    
    #####################################################################################
    # Pilot-Job State
    def set_pilot_description(self, pilot_url, description):     
        logger.debug("update description of pilot job to: " + str(description))
        self.redis_client.hmset(pilot_url + ":description", {"description":description})
        
    def get_pilot_description(self, pilot_url):
        description = self.redis_client.hgetall(pilot_url + ":description")
        return description
    
    #def is_pilot_stopped(self,pilot_url):
    #    state = self.redis_client.hgetall(pilot_url)
    #    if state==None or not state.has_key("stopped"):
    #        return True        
    #    return state["stopped"]
    
    def get_jobs_of_pilot(self, pilot_url):
        """ returns array of job_url that are associated with a pilot """
        jobs = self.redis_client.keys(pilot_url+":jobs:*")
        jobs_fqdn = [os.path.join(self.get_address(), i)for i in jobs] 
        return jobs_fqdn
    
    
    def delete_pilot(self, pilot_url):
        items = self.redis_client.keys(pilot_url+"*")  
        for i in items:        
            self.pipe.delete(i)
        self.pipe.execute()
    
    #####################################################################################
    # Sub-Job State    
    def set_job_state(self, job_url, new_state):
        #self.resource_lock.acquire()        
        logger.debug("set job state to: " + str(new_state))
        self.redis_client.hset(job_url, "state", str(new_state))
        
        if new_state=="Unknown":
            self.redis_client.hset(job_url,"start_time", str(time.time()))
        elif new_state=="Running":
            self.redis_client.hset(job_url,"end_queue_time", str(time.time()))
        elif new_state=="Done":
            self.redis_client.hset(job_url, "end_time", str(time.time()))
       
        
        #self.resource_lock.release()
        
    def get_job_state(self, job_url):
        return self.redis_client.hget(job_url, "state")      
    
    
    #####################################################################################
    # Sub-Job Description
    def set_job(self, job_url, job_dict):
        self.redis_client.hmset(job_url, job_dict)
        self.set_job_state(job_url, "Unknown")
        
    def get_job(self, job_url):
        return self.redis_client.hgetall(job_url)    
    
    def delete_job(self, job_url):
        self.redis_client.delete(job_url+"*")
    
    
    #####################################################################################
    # Distributed queue for sub-jobs
    def queue_job(self, pilot_url, job_url):
        """ queue new job to pilot """
        queue_name = pilot_url + ":queue"
        self.redis_client.set(queue_name + ':last_in', pickle.dumps(datetime.datetime.now()))
        self.redis_client.lpush(queue_name, job_url)
                
        
    def dequeue_job(self, pilot_url):
        """ deque to new job  of a certain pilot """
        queue_name = pilot_url + ":queue"        
        logger.debug("Dequeue sub-job from: " + queue_name 
                      + " number queued items: " + str(self.redis_client.llen(queue_name)))
        self.redis_client.set(queue_name + ':last_out', pickle.dumps(datetime.datetime.now()))
        job_url = self.redis_client.brpop(queue_name, 1)
        if job_url==None:
            return job_url
        logger.debug("Dequeued: " + str(job_url))
        return job_url[1]
    



'''
Amazon S3 based File Transfer Implementation
'''
import urlparse

import errno
import sys
import os
import stat
import logging
import traceback
import time

# This is for local debugging!
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../.."))

import bliss.saga as saga
from pilot.api import State
from bigjob import logger

##################

from boto.s3.connection import S3Connection, OrdinaryCallingFormat
from boto.s3.key import Key
from boto.s3.connection import Location

# Authentication
# Please use ~/.boto file to configure your security credentials (if possible)
# see http://boto.readthedocs.org/en/latest/boto_config_tut.html
# 
# [Credentials]
# aws_access_key_id = <your access key>
# aws_secret_access_key = <your secret key>
#
# Alternatively you can use these two variables
AWS_ACCESS_KEY_ID=None
AWS_SECRET_ACCESS_KEY=None

class S3FileAdaptor(object):
    """ BigData File Management for Pilot Data 
        Supports pilot data on top of S3 and Eucalyptus Walrus
        
            s3://
            walrus://<endpoint-ip>    
            
        Amazon S3 supported regions:
        
            Default is an empty string "" => us-east-1
        
            'ap-northeast-1'
            'ap-southeast-1'
            'ap-southeast-2'
            'EU'
            'sa-east-1'
            'us-west-1'
            'us-west-2'
        
    """
    
   
    
    def __init__(self, resource_url, security_context=None, pilot_data_description=None):        
        self.resource_url = saga.Url(resource_url)
        self.bucket_name = self.__get_bucket_name(resource_url)
        self.__state=State.New
        self.pilot_data_description = pilot_data_description
        aws_access_key_id=None
        aws_secret_access_key=None
        
        # try to recover key from pilot_data_description
        if self.pilot_data_description!=None and\
           self.pilot_data_description.has_key("access_key_id") and \
           self.pilot_data_description.has_key("secret_access_key"):
            aws_access_key_id=self.pilot_data_description["access_key_id"]
            aws_secret_access_key=self.pilot_data_description["secret_access_key"]
        
        # try to recover key from security context
        if security_context!=None:
            logger.debug("Attempt to restore credentials from security context: " + str(security_context))
            security_context = eval(security_context)
            aws_access_key_id=security_context["aws_access_key_id"]
            aws_secret_access_key=security_context["aws_secret_access_key"]
    
        self.s3_conn=None
        if self.resource_url.scheme == "walrus" or self.resource_url.scheme == "swift":
            calling_format=OrdinaryCallingFormat()
            logger.debug("Access Key: %s Secret: %s Host: %s"%(aws_access_key_id,
                                                               aws_secret_access_key,
                                                               self.resource_url.host)
                         )
            port = 8773
            if self.resource_url.port!=None:
                port = self.resource_url.port
            
            path = "/"
            if self.resource_url.scheme == "walrus":
                path = "/services/Walrus"
                
            self.s3_conn = S3Connection(aws_access_key_id=aws_access_key_id,
                                        aws_secret_access_key=aws_secret_access_key,
                                        is_secure=False,
                                        host=self.resource_url.host,
                                        port=port,
                                        calling_format=calling_format,
                                        path=path)
        else: # s3:// urls
            self.s3_region = None
            # Region specifier according to Amazon API:
            # http://docs.aws.amazon.com/AmazonS3/latest/API/RESTBucketGETlocation.html
            if self.pilot_data_description.has_key("region"):
                self.s3_region =  self.pilot_data_description["region"]
           
            self.s3_conn = S3Connection(
                                        aws_access_key_id, 
                                        aws_secret_access_key, 
                                        )
      
    
    
    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        return {"aws_access_key_id": self.s3_conn.aws_access_key_id,
                "aws_secret_access_key": self.s3_conn.aws_secret_access_key}

                       
        
    def initialize_pilotdata(self):
        # Create bucket
        try:
            if self.s3_region==None:
                logger.debug("Use default S3 region.")
                self.s3_region = "" # Default for US East
            self.bucket = self.s3_conn.create_bucket(self.bucket_name, location=self.s3_region)
        except:
            # bucket already exists
            #self.__print_traceback()
            self.bucket = self.s3_conn.get_bucket(self.bucket_name)
            
        self.__state=State.Running
       
        
    def get_pilotdata_size(self):
        # unlimited size
        return None
    
    
    def delete_pilotdata(self):
        self.__state=State.Done
    
        
    def get_state(self):
        return self.__state
        
            
    def create_du(self, du_id):
        logger.debug("create object: " + du_id)
        k = Key(self.bucket)
        k.key = str(du_id)+"/du_info"
        k.set_contents_from_string(du_id)
                 
                 
    def put_du(self, du):
        logger.debug("Copy DU to S3/Walrus")
        du_items = du.list()
        for i in du_items.keys():     
            try:
                local_filename=du_items[i]["local"]
                remote_path = os.path.join(str(du.id), os.path.basename(local_filename))
                logger.debug("copy %s to %s"%(local_filename, remote_path))
                self._put_file(local_filename, remote_path)
            except:
                logger.debug("Could not copy: " + str(i))
                
    
    def get_du(self, du, target_url):
        #du_id = "du-7370d7b5-ed0b-11e1-95df-705681b3df0f"
        du_id = du.id
        logger.debug("Get DU: " + str(du_id))
        result = self.bucket.list(prefix=du_id)
        logger.debug("Result Bucket List: " + str(result))
        for key in result:
            logger.debug(str(key))
            full_filename = key.name
            if full_filename != None:
                logger.debug("Process file: " + full_filename)
                if not full_filename.endswith("/"):
                    self._get_file(full_filename, os.path.join(target_url, os.path.basename(full_filename)))
        
   
    def copy_du(self, du, pd_new):
        remote_url = pd_new.resource_url + "/" + str(du.id)
        local_url =  self.resource_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
    
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
    
    ###########################################################################
    # Pure File Management APIs
    def _put_file(self, source, target):
        logger.debug("Put file: %s to %s"%(source, target))
        if self.__starts_with_valid_prefix(source):
            logger.debug("Copy file from S3/Walrus")
            source_bucket_name = self.__get_bucket_name(source)
            source_key_name = self.__get_key_name(source)
            self.bucket.copy_key(target, source_bucket_name, source_key_name)
            #k = Key(source_bucket_name)
            #k.copy(self.bucket_name, target)
        else:
            logger.debug("Copy file from Local")
            k = Key(self.bucket)
            k.key=target
            k.set_contents_from_filename(source)
            logger.debug("Put file result: %s"%source)
    
    
    def _get_file(self, source, target):
        logger.debug("GET file: %s to %s"%(source, target))
        k = self.bucket.get_key(source)
        k.key=source
        k.get_contents_to_filename(target)
        
         
    def transfer(self, source_url, target_url):
        pass
    
    def create_remote_directory(self, target_url):
        return True
    
                   
    ###########################################################################
    def __starts_with_valid_prefix(self, url):
        valid_prefix=["s3", "walrus"]
        result = False
        for i in valid_prefix:
            result = url.startswith(i)
            if result == True:
                break
        return result
    
    def __get_bucket_name(self, resource_url):
        surl = saga.Url(resource_url)
        if surl.scheme.startswith("s3"):
            bucket_name = resource_url.replace("s3://", "")
            try:
                bucket_name = bucket_name[:bucket_name.index("/")]
            except:
                pass
            #bucket_name = bucket_name.replace("/", "")
        else:
            bucket_name = surl.path[1:]
        return bucket_name
        
   
    def __get_key_name(self, resource_url):
        surl = saga.Url(resource_url)
        # get path out of URL
        if surl.scheme.startswith("s3"):
            bucket_name = resource_url.replace("s3://", "")            
        else:
            bucket_name = surl.path[1:]
        
        # get key path out of URL
        try:
            key_name = bucket_name[bucket_name.index("/")+1:]
        except:
            pass
        
        return key_name
    
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
    
def test_walrus():
    s3 = S3FileAdaptor("walrus://149.165.146.135/pilot-data-c4eb26eb-ed0c-11e1-ac98-705681b3df0f", 
                       pilot_data_description={ "access_key_id":"8MCXRAMXMHDYKWNKXZ8WF",
                                                "secret_access_key":"YrcUqSw2Arxshrh3ZtenkxerWwCWdMTKvZYoLPAo" })
    s3.initialize_pilotdata()
    s3._put_file("test.txt", "du-7370d7b5-ed0b-11e1-95df-705681b3df0f/test.txt")
    s3._get_file("du-7370d7b5-ed0b-11e1-95df-705681b3df0f/test.txt", "test2.txt")
    s3.get_du("du-7370d7b5-ed0b-11e1-95df-705681b3df0f", ".")
    
def test_swift():
    s3 = S3FileAdaptor("swift://149.165.146.50:3333/pilot-data-c4eb26eb-ed0c-11e1-ac98-705681b3df0f", 
                       pilot_data_description={ "access_key_id":"f9716a49c92a4a4cbedb6aba5e78d682",
                                                "secret_access_key":"bcdff54b7fe94d63b4412c762e823a84" })
    s3.initialize_pilotdata()
    s3._put_file("test.txt", "du-7370d7b5-ed0b-11e1-95df-705681b3df0f/test.txt")
    s3._get_file("du-7370d7b5-ed0b-11e1-95df-705681b3df0f/test.txt", "test2.txt")
    s3.get_du("du-7370d7b5-ed0b-11e1-95df-705681b3df0f", ".")
   
def test_s3import():
    s3 = S3FileAdaptor("s3://pilot-data-andre-test-create-from-s3-url", 
                       pilot_data_description={ "access_key_id":"AKIAJPGNDJRYIG5LIEUA",
                                                "secret_access_key":"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" })
    s3.initialize_pilotdata()
    s3._put_file("s3://pilot-data-05d88e40-f65b-11e1-a327-00215ec9e3ac/du-3624837e-f66f-11e1-a327-00215ec9e3ac/WRT54GS_UG_WEB_20070529.pdf", "bla/test.pdf")

def test_s3import_via_pilotapi():
    COORDINATION_URL="redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379"
    from pilot import PilotComputeService, PilotDataService, ComputeDataService, State
    pilot_data_service = PilotDataService(coordination_url=COORDINATION_URL)
    
    ###################################################################################################
    # Pick one of the Pilot Data Descriptions below    
    
    pilot_data_description_aws={
                                "service_url": "s3://pilot-data-andre-workflow",
                                "size": 100,   
                                "affinity_datacenter_label": "us-east-1",              
                                "affinity_machine_label": ""    ,
                                "access_key_id": "AKIAJPGNDJRYIG5LIEUA",
                                "secret_access_key":"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",                                                         
                                }

    pd = pilot_data_service.create_pilot(pilot_data_description=pilot_data_description_aws)
     
    data_unit_description = {
                              "file_urls": ['s3://pilot-data-cec5d816-fa8f-11e1-ab5e-e61f1322a75c/du-67b4c762-fa90-11e1-ab5e-e61f1322a75c/ip-10-84-173-21512MB_2.input-chunk-02'],
                              "affinity_datacenter_label": "us-east-1",              
                              "affinity_machine_label": ""
                             }    
      
    # submit pilot data to a pilot store 
    input_data_unit = pd.submit_data_unit(data_unit_description)
    input_data_unit.wait()
    

if __name__ == "__main__":
    test_s3import_via_pilotapi()
    


'''
Encapsulates coordination and communication specifics of bigjob
'''

import logging
import threading
import datetime
import sys
import os
import pickle
import pdb
import zmq
import traceback
import Queue
import socket
import time
from zmq.eventloop import ioloop, zmqstream
import zlib, cPickle as pickle
from bigjob import logger

if sys.version_info < (2, 5):
    sys.path.append(os.path.dirname( os.path.abspath( __file__) ) + "/../ext/uuid-1.30/")
    sys.stderr.write("Warning: Using unsupported Python version\n")
    
logging.debug(str(sys.path))
import uuid

SERVER_IP="localhost"
SERVER_PORT=0

NUMBER_RETRIES=2

class message:    
    def __init__(self, command, key, value):
        self.command = command
        self.key = key
        self.value = value

    def __repr__(self):
        return ("command: %s, key: %s, value: %s "%(self.command, self.key, self.value))

class bigjob_coordination(object):
    '''
    Encapsulates communication and coordination
    Implementation based on ZMQ 
    '''
    def __init__(self, server=SERVER_IP, server_port=SERVER_PORT, server_connect_url=None,
                 username=None, password=None, dbtype=None, url_prefix=None):
        '''
        Constructor
        set server and server_port to create a service (server)
        set server_connect_url to connect to a service (client)
        '''  
        self.stopped = False
        self.has_stopped=False        
        
        
        # state managed by server
        self.pilot_states = {}
        self.job_ids = []
        self.jobs = {}
        self.job_states = {}
        self.new_job_queue = Queue.Queue()
        
        # Lock for server and client to manage concurrent access
        self.resource_lock = threading.Lock()
        
        # Client side queue
        self.subjob_queue = Queue.Queue()
            
        # set up ZMQ client / server communication
        self.context = zmq.Context()
       
        self.server_role = False  
        self.address = None 
        self.dbtype = ""   
        if server == "*":
            server = socket.gethostname()        
        logging.debug("Server: " + server)         
        if server_connect_url==None: # role = Server
            self.server_role = True
            # start eventloop
            self.startup_condition = threading.Condition()
            self.eventloop_thread=threading.Thread(target=self.__server, args=(server, server_port))
            self.eventloop_thread.daemon=True
            self.eventloop_thread.start()
            
            logging.debug("Setting up socket for notifications")
            # socket for sending notification
            self.push_socket = self.context.socket(zmq.PUSH)
            push_port = self.push_socket.bind_to_random_port("tcp://*")    
            self.push_address = "tcp://"+server+":"+str(push_port)                
            
            
            logging.debug("Waiting for server to complete startup")
            # wait for server thread to complete startup
            self.startup_condition.acquire()
            while self.address == None:
                self.startup_condition.wait()
            self.startup_condition.release()                       
        else: # role client
            urls = server_connect_url.split(",")
            self.address = urls[0]
            self.push_address = urls[1]           
            self.server_role = False
            self.pilot_url = server_connect_url
            
           
            
        
        logging.debug("Connect sockets to server: " + self.address + " push: " + self.push_address)
        # connect to REP server
        self.client_socket = self.context.socket(zmq.REQ)
        self.client_socket.connect(self.address)
        
        # connect to PUSH server
        self.pull_socket = self.context.socket(zmq.PULL)
        self.pull_socket.connect(self.push_address)
        
        if self.server_role==False:
            self.notification_thread=threading.Thread(target=self.__wait_for_notifications)
            self.notification_thread.daemon=True
            self.notification_thread.start()
        
        
        logging.debug("Connected to REP socket at: " + self.address + " and PUSH socket at: " + self.push_address)
        logging.debug("C&C ZMQ system initialized")
        

        
    def get_address(self):
        """ return handle to c&c subsystems """
        return self.address+"," +self.push_address
        
    #####################################################################################
    # Pilot-Job State
    def set_pilot_state(self, pilot_url, new_state, stopped=False):     
        logging.debug("BEGIN update state of pilot job to: " + str(new_state) 
                       + " Lock: " + str(self.resource_lock))
        counter = 0
        result = None
        while result != "SUCCESS" and counter < NUMBER_RETRIES:
            with self.resource_lock:  
                msg = message("set_pilot_state", pilot_url, {"state":str(new_state), "stopped":str(stopped)})
                try:
                    self.client_socket.send_pyobj(msg, zmq.NOBLOCK)        
                    result = self.client_socket.recv_pyobj()
                except:
                    traceback.print_exc(file=sys.stderr)
            # stop background thread running the server (if True)
            self.stopped=stopped             
            if result == None:
                logging.error("RETRY set_pilot_state")
                counter = counter + 1
                time.sleep(1) 
        logging.debug("END update state of pilot job to: " + str(new_state))
        
    def get_pilot_state(self, pilot_url):
        logging.debug("BEGIN get_pilot_state: %s lock: %s" % (pilot_url, str(self.resource_lock)))
        counter = 0
        result = None
        while result ==None and counter < NUMBER_RETRIES:
            with self.resource_lock:   
                msg = message("get_pilot_state", pilot_url, " Lock: ")
                try:
                    self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
                    result = self.client_socket.recv_pyobj()
                except:
                    pass
                  
            logging.debug("END get_pilot_state: %s state: %s, lock: %s" % 
                          (pilot_url, str(result.value), str(self.resource_lock)))
            if result == None:               
                counter = counter + 1
                logging.error("RETRY get_pilot_state - Retry # %d"%counter)
                time.sleep(1) 
        return result.value
        logging.debug("END get_pilot_state: %s lock: %s" % (pilot_url, str(self.resource_lock)))
    
    def get_jobs_of_pilot(self, pilot_url):
        """ returns array of job_url that are associated with a pilot """
        """ local only - used only by manager """
        return self.job_ids
    
    def delete_pilot(self, pilot_url):
        """ local only - used only by manager """
        # stop everything
        self.stopped=True
        msg = message("STOP", pilot_url, "")        
        self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
        if self.server_role == True:
            self.push_socket.send_pyobj(msg, zmq.NOBLOCK)
        #self.eventloop_thread.join()
        logging.debug("Has stopped: " + str(self.has_stopped))
        self.__shutdown()
        
    
    #####################################################################################
    # Sub-Job State    
    def set_job_state(self, job_url, new_state):
        logging.debug("Set job state: %s to %s"%(job_url, new_state))
        counter = 0
        result = None
        while result == None and counter < NUMBER_RETRIES:
            with self.resource_lock:   
                msg = message("set_job_state", job_url, new_state)
                try:
                    self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
                    result = self.client_socket.recv_pyobj()
                except:
                    traceback.print_exc(file=sys.stderr)
      
            if result == None:
                counter = counter + 1
                logging.error("RETRY %d set_job_state %s to %s"%(counter, job_url, new_state))
                if counter == NUMBER_RETRIES-1:
                    self.__reset_client_socket()              
                time.sleep(2)
                continue # retry
            else:
                logging.debug("SUCCESS set_job_state (%s to %s)"%(job_url, new_state))
            
        
        
    def get_job_state(self, job_url):
        #logging.debug("get_job_state")
        with self.resource_lock:   
            msg = message("get_job_state", job_url, "")
            self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
            result = self.client_socket.recv_pyobj()            
        return result.value      
    
    #####################################################################################
    # Pilot-Job Description
    def set_pilot_description(self, pilot_url, description):     
        pass
    
    def get_pilot_description(self, pilot_url):
        pass
    
    #####################################################################################
    # Sub-Job Description
    def set_job(self, job_url, job_dict):        
        """ local only - used only by manager """
        self.job_ids.append(job_url)
        self.jobs[job_url] = job_dict   
        self.job_states[job_url]="Unknown"     
    
    def get_job(self, job_url):       
        if self.jobs.has_key(job_url)==False:
            logging.debug("get_job: " + str(self.resource_lock))
            with self.resource_lock: 
                logging.debug("get_job (lock acquired): " + str(self.resource_lock))
                msg = message("get_job", job_url, "")
                self.client_socket.send_pyobj(msg)
                result = self.client_socket.recv_pyobj()
                self.jobs[job_url] = result.value
                logging.debug("received job: "  + str(result.value))
        return self.jobs[job_url] 
    
    def delete_job(self, job_url):
        self.job_ids.remove(job_url)
        del self.jobs[job_url]
        del self.job_ids[job_url]
    
    
    #####################################################################################
    # Distributed queue for sub-jobs
    def queue_job(self, pilot_url, job_url):        
        if self.server_role == False: # just re-queue locally at client
            self.subjob_queue.put(job_url)        
        elif self.server_role == True:
            """ queue new job to pilot """
            logging.debug("queue_job " + str(self.resource_lock))
            counter = 0
            result = None
            success = False
            while result ==None and counter < NUMBER_RETRIES:
                with self.resource_lock:   
                    msg = message("queue_job", "", job_url)
                    try:
                        self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
                        result = self.client_socket.recv_pyobj()
                        success=True
                    except:
                        traceback.print_exc(file=sys.stderr)     
                if result == None:
                    counter = counter + 1
                    logging.error("RETRY %d queue_job"%counter)
                    if counter == NUMBER_RETRIES and success==False:
                        self.__reset_client_socket()
                    time.sleep(2)
            msg2 = message("notification", "", job_url)
            self.push_socket.send_pyobj(msg2)
        
        return success             
             
        
    def dequeue_job(self, pilot_url):
        """ dequeue to new job  of a certain pilot """
        return self.subjob_queue.get()
        
    
    
    
    #####################################################################################
    # Private functions    
    def __server_handler(self, messageList):
        """ server for managing job / pilot job states via ZMQ """
        msg = pickle.loads(messageList[0])
        logging.debug("Message received: " + str(msg))
        self.__handle_message(msg, self.stream)
        
    
    def __handle_message(self, msg, reply_socket):
        try:
            command = msg.command        
            if command == "set_pilot_state":
                self.pilot_states[msg.key] = msg.value
                reply_socket.send_pyobj("SUCCESS")
                #self.service_socket.send("")            
            elif command == "get_pilot_state":
                result = message ("", "", self.pilot_states[msg.key])
                reply_socket.send_pyobj(result, zmq.NOBLOCK)                
            elif command == "set_job_state":
                self.job_states[msg.key] = msg.value
                reply_socket.send_pyobj("SUCCESS", zmq.NOBLOCK)       
            elif command == "get_job_state":
                result=message("", "", self.job_states[msg.key])
                reply_socket.send_pyobj(result, zmq.NOBLOCK)            
            elif command == "get_job":
                result = message("","", self.jobs[msg.key])
                reply_socket.send_pyobj(result, zmq.NOBLOCK)
            elif command == "queue_job":                
                self.new_job_queue.put(msg.value)
                reply_socket.send_pyobj("SUCCESS", zmq.NOBLOCK)                       
            elif command == "dequeue_job":
                new_job=None
                try:
                    new_job = self.new_job_queue.get(False)
                except:                
                    pass
                result = message("","", new_job)
                reply_socket.send_pyobj(result, zmq.NOBLOCK)
            else:
                logging.debug("sending default reply")
                reply_socket.send_pyobj("", zmq.NOBLOCK)
        except:
            traceback.print_exc(file=sys.stderr)
    
    
    def __server(self, server, server_port):
        """ server for managing job / pilot job states via ZMQ """
        service_socket = self.context.socket(zmq.REP)
        if SERVER_PORT==0: # random port
            server_port = service_socket.bind_to_random_port("tcp://*")    
            self.address = "tcp://"+server+":"+str(server_port)                
        elif server == "localhost":
            self.server_address = "tcp://*:"+str(server_port)
            self.address = "tcp://"+server+":"+str(server_port)
            service_socket.bind(self.server_address)
        else:
            self.server_address = "tcp://"+server+":"+str(server_port)
            self.address = self.server_address
            service_socket.bind(self.server_address)
        logging.debug("Starting service at: " + self.address)        
        self.startup_condition.acquire()
        self.startup_condition.notifyAll()   
        self.startup_condition.release()
        logging.debug("Startup condition signaled")
        while self.stopped == False:
            #logging.debug("Waiting for messages...")
            try:
                msg = service_socket.recv_pyobj()            
                #logging.debug("Message received: " + str(msg))
                self.__handle_message(msg, service_socket)
            except:
                pass
            #logging.debug("Message handled: " + str(msg) + " stopped = " + str(self.stopped))
            #pdb.set_trace()
        logging.debug("__server thread stopped: " + str(self.stopped))
        self.has_stopped = True
        #service_socket.close()
            
    def __loop(self):
        self.loop = ioloop.IOLoop.instance()
        #self.loop.add_handler(self.service_socket, self.__server_handler, zmq.POLLIN)
        self.stream = zmqstream.ZMQStream(self.service_socket, self.loop)
        self.stream.on_recv(self.__server_handler)
        logging.debug("Start event loop")
        self.loop.start()
        
    def __reset_client_socket(self):
        logging.error("RESETING client socket")
        with self.resource_lock: 
            try:
                self.client_socket.close()
            except:
                traceback.print_exc(file=sys.stderr)
            self.client_socket = self.context.socket(zmq.REQ)
            self.client_socket.connect(self.address)
        
    def __wait_for_notifications(self):
        """ waits for notifications and puts new jobs into queue """    
        #while result==None and counter < NUMBER_RETRIES:
        while self.stopped == False:
            # read object from queue
            logging.debug(" __wait_for_notifications: polling for new jobs - stopped: " + str(self.stopped))
            msg = message ("dequeue_job", self.pilot_url, "")
            try:
                with self.resource_lock: 
                    self.client_socket.send_pyobj(msg, zmq.NOBLOCK)
                    result = self.client_socket.recv_pyobj().value
                logging.debug(" __wait_for_notifications: received new jobs " + str(result))
                if result != None:
                    self.subjob_queue.put(result)
                    time.sleep(0.2)
                    continue                                              
            except:                
                traceback.print_exc(file=sys.stderr)                    
                logging.error("Error dequeuing job")                  
                time.sleep(1)
                continue   
            logging.debug(" __wait_for_notifications: End Loop - stopped " + str(self.stopped))
                
            #if counter == NUMBER_RETRIES-1 and success == False:
            #    self.__reset_client_socket()              
            #    time.sleep(2)
            #    continue # retry
            
            #if counter == NUMBER_RETRIES and success == False:
            #    return result
            
            logging.debug(" __wait_for_notifications: wait for notification")
            # wait for next job notification
            if result == None:
                try:
                    logging.debug("wait for notification")
                    self.pull_socket.recv_pyobj()
                    logging.debug("received notification")
                except:
                    pass
        
        # wait for next job notification
        #while self.stopped == False:
        #    logging.debug("wait for notification")
        #    self.pull_socket.recv_pyobj()
        #    logging.debug("received notification")
            
    def __shutdown(self):
        logging.debug("shutdown ZMQ")
        try:     
            #self.client_socket.close()            
            #self.service_socket.close()
            #self.context.term()
            pass
        except:
            pass       
            
            


import sys
import os
import time

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
sys.path.insert(0, os.getcwd() + "/../")
from pilot import PilotComputeService, ComputeDataServiceDecentral, State

COORDINATION_URL = "redis://localhost:6379"

if __name__ == "__main__":      
    
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description = {
                             "service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             "working_directory": os.path.join(os.getcwd(),"work"),
                             'affinity_datacenter_label': "eu-de-south",              
                             'affinity_machine_label': "mymachine" 
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
    pilotjob2 = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
    
    compute_data_service = ComputeDataServiceDecentral()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    
    # start work unit
    compute_unit_description = {
            "executable": "/bin/date",
            "arguments": [""],
            "number_of_processes": 1,            
            "output": "stdout.txt",
            "error": "stderr.txt",
    }    
    
    for i in range(0,4):
        compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
    
    
    print("Finished setup. Waiting for scheduling of CU")
    compute_data_service.wait()
    
    while compute_unit != State.Done:
        print("Final state check...")
        state_cu = compute_unit.get_state()
        print "PCS State %s" % pilot_compute_service
        print "CU: %s State: %s"%(compute_unit, state_cu)
        if state_cu==State.Done:
            break
        time.sleep(2)  
    
    print("Terminate Pilot Compute and Compute Data Service")
    compute_data_service.cancel()    
    pilot_compute_service.cancel()
